package org.antlr.analysis;

package .var.folders.cl.9twcgjfx4vsg5zl73c6wxmvm0000gn.T;

import java.util.ArrayList;
import java.util.Collections;
import java.util.Comparator;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;
import java.util.Set;
import org.antlr.analysis.DFA;
import org.antlr.analysis.DFAOptimizer;
import org.antlr.analysis.DFAState;
import org.antlr.analysis.Label;
import org.antlr.analysis.NFAConfiguration;
import org.antlr.analysis.NFAContext;
import org.antlr.analysis.NFAState;
import org.antlr.analysis.NonLLStarDecisionException;
import org.antlr.analysis.PredicateLabel;
import org.antlr.analysis.RuleClosureTransition;
import org.antlr.analysis.SemanticContext;
import org.antlr.analysis.Transition;
import org.antlr.misc.BitSet;
import org.antlr.misc.OrderedHashSet;
import org.antlr.misc.Utils;
import org.antlr.runtime.Token;
import org.antlr.tool.ErrorManager;

public class NFAToDFAConverter {
  protected List<DFAState> work = new LinkedList<DFAState>();
  
  protected NFAContext[] contextTrees;
  
  protected DFA dfa;
  
  public static boolean debug = false;
  
  public static boolean SINGLE_THREADED_NFA_CONVERSION = true;
  
  protected boolean computingStartState = false;
  
  public NFAToDFAConverter(DFA dfa) {
    this.dfa = dfa;
    int nAlts = dfa.getNumberOfAlts();
    initContextTrees(nAlts);
  }
  
  public void convert() {
    this.dfa.startState = computeStartState();
    while (this.work.size() > 0 && !this.dfa.nfa.grammar.NFAToDFAConversionExternallyAborted()) {
      DFAState d = this.work.get(0);
      if (this.dfa.nfa.grammar.composite.watchNFAConversion)
        System.out.println("convert DFA state " + d.stateNumber + " (" + d.nfaConfigurations.size() + " nfa states)"); 
      int k = this.dfa.getUserMaxLookahead();
      if (k > 0 && k == d.getLookaheadDepth()) {
        resolveNonDeterminisms(d);
        if (d.isResolvedWithPredicates()) {
          addPredicateTransitions(d);
        } else {
          d.setAcceptState(true);
        } 
      } else {
        findNewDFAStatesAndAddDFATransitions(d);
      } 
      this.work.remove(0);
    } 
    this.dfa.findAllGatedSynPredsUsedInDFAAcceptStates();
  }
  
  protected DFAState computeStartState() {
    NFAState alt = this.dfa.decisionNFAStartState;
    DFAState startState = this.dfa.newState();
    this.computingStartState = true;
    int i = 0;
    int altNum = 1;
    while (alt != null) {
      NFAContext initialContext = this.contextTrees[i];
      if (i == 0 && (this.dfa.getNFADecisionStartState()).decisionStateType == 1) {
        int numAltsIncludingExitBranch = this.dfa.nfa.grammar.getNumberOfAltsForDecisionNFA(this.dfa.decisionNFAStartState);
        altNum = numAltsIncludingExitBranch;
        closure((NFAState)(alt.transition[0]).target, altNum, initialContext, SemanticContext.EMPTY_SEMANTIC_CONTEXT, startState, true);
        altNum = 1;
      } else {
        closure((NFAState)(alt.transition[0]).target, altNum, initialContext, SemanticContext.EMPTY_SEMANTIC_CONTEXT, startState, true);
        altNum++;
      } 
      i++;
      if (alt.transition[1] == null)
        break; 
      alt = (NFAState)(alt.transition[1]).target;
    } 
    this.dfa.addState(startState);
    this.work.add(startState);
    this.computingStartState = false;
    return startState;
  }
  
  protected void findNewDFAStatesAndAddDFATransitions(DFAState d) {
    OrderedHashSet<Label> labels = d.getReachableLabels();
    Label EOTLabel = new Label(-2);
    boolean containsEOT = (labels != null && labels.contains(EOTLabel));
    if (!this.dfa.isGreedy() && containsEOT) {
      convertToEOTAcceptState(d);
      return;
    } 
    int numberOfEdgesEmanating = 0;
    Map<Integer, Transition> targetToLabelMap = new HashMap<Integer, Transition>();
    int numLabels = 0;
    if (labels != null)
      numLabels = labels.size(); 
    for (int i = 0; i < numLabels; i++) {
      Label label = labels.get(i);
      DFAState t = reach(d, label);
      if (debug)
        System.out.println("DFA state after reach " + label + " " + d + "-" + label.toString(this.dfa.nfa.grammar) + "->" + t); 
      if (t != null) {
        if (t.getUniqueAlt() == -1)
          closure(t); 
        DFAState targetState = addDFAStateToWorkList(t);
        numberOfEdgesEmanating += addTransition(d, label, targetState, targetToLabelMap);
        targetState.setLookaheadDepth(d.getLookaheadDepth() + 1);
      } 
    } 
    if (!d.isResolvedWithPredicates() && numberOfEdgesEmanating == 0) {
      this.dfa.probe.reportDanglingState(d);
      int minAlt = resolveByPickingMinAlt(d, null);
      d.setAcceptState(true);
      this.dfa.setAcceptState(minAlt, d);
    } 
    if (d.isResolvedWithPredicates())
      addPredicateTransitions(d); 
  }
  
  protected static int addTransition(DFAState d, Label label, DFAState targetState, Map<Integer, Transition> targetToLabelMap) {
    int n = 0;
    if (DFAOptimizer.COLLAPSE_ALL_PARALLEL_EDGES) {
      Integer tI = Utils.integer(targetState.stateNumber);
      Transition oldTransition = targetToLabelMap.get(tI);
      if (oldTransition != null) {
        if (label.getAtom() == -2) {
          oldTransition.label = new Label(-2);
        } else if (oldTransition.label.getAtom() != -2) {
          oldTransition.label.add(label);
        } 
      } else {
        n = 1;
        label = (Label)label.clone();
        int transitionIndex = d.addTransition(targetState, label);
        Transition trans = d.getTransition(transitionIndex);
        targetToLabelMap.put(tI, trans);
      } 
    } else {
      n = 1;
      d.addTransition(targetState, label);
    } 
    return n;
  }
  
  public void closure(DFAState d) {
    if (debug)
      System.out.println("closure(" + d + ")"); 
    List<NFAConfiguration> configs = new ArrayList<NFAConfiguration>();
    configs.addAll(d.nfaConfigurations);
    int numConfigs = configs.size();
    for (int i = 0; i < numConfigs; i++) {
      NFAConfiguration c = configs.get(i);
      if (!c.singleAtomTransitionEmanating)
        closure(this.dfa.nfa.getState(c.state), c.alt, c.context, c.semanticContext, d, false); 
    } 
    d.closureBusy = null;
  }
  
  public void closure(NFAState p, int alt, NFAContext context, SemanticContext semanticContext, DFAState d, boolean collectPredicates) {
    if (debug)
      System.out.println("closure at " + p.enclosingRule.name + " state " + p.stateNumber + "|" + alt + " filling DFA state " + d.stateNumber + " with context " + context); 
    NFAConfiguration proposedNFAConfiguration = new NFAConfiguration(p.stateNumber, alt, context, semanticContext);
    if (closureIsBusy(d, proposedNFAConfiguration)) {
      if (debug) {
        System.out.println("avoid visiting exact closure computation NFA config: " + proposedNFAConfiguration + " in " + p.enclosingRule.name);
        System.out.println("state is " + d.dfa.decisionNumber + "." + d.stateNumber);
      } 
      return;
    } 
    d.closureBusy.add(proposedNFAConfiguration);
    d.addNFAConfiguration(p, proposedNFAConfiguration);
    Transition transition0 = p.transition[0];
    if (transition0 instanceof RuleClosureTransition) {
      int depth = context.recursionDepthEmanatingFromState(p.stateNumber);
      if (depth == 1 && d.dfa.getUserMaxLookahead() == 0) {
        d.dfa.recursiveAltSet.add(alt);
        if (d.dfa.recursiveAltSet.size() > 1) {
          d.abortedDueToMultipleRecursiveAlts = true;
          throw new NonLLStarDecisionException(d.dfa);
        } 
      } 
      if (depth >= NFAContext.MAX_SAME_RULE_INVOCATIONS_PER_NFA_CONFIG_STACK) {
        d.abortedDueToRecursionOverflow = true;
        d.dfa.probe.reportRecursionOverflow(d, proposedNFAConfiguration);
        if (debug)
          System.out.println("analysis overflow in closure(" + d.stateNumber + ")"); 
        return;
      } 
      RuleClosureTransition ref = (RuleClosureTransition)transition0;
      NFAContext newContext = new NFAContext(context, p);
      NFAState ruleTarget = (NFAState)ref.target;
      closure(ruleTarget, alt, newContext, semanticContext, d, collectPredicates);
    } else if (p.isAcceptState() && context.parent != null) {
      NFAState whichStateInvokedRule = context.invokingState;
      RuleClosureTransition edgeToRule = (RuleClosureTransition)whichStateInvokedRule.transition[0];
      NFAState continueState = edgeToRule.followState;
      NFAContext newContext = context.parent;
      closure(continueState, alt, newContext, semanticContext, d, collectPredicates);
    } else {
      if (transition0 != null && transition0.isEpsilon()) {
        boolean collectPredicatesAfterAction = collectPredicates;
        if (transition0.isAction() && collectPredicates)
          collectPredicatesAfterAction = false; 
        closure((NFAState)transition0.target, alt, context, semanticContext, d, collectPredicatesAfterAction);
      } else if (transition0 != null && transition0.isSemanticPredicate()) {
        SemanticContext labelContext = transition0.label.getSemanticContext();
        if (this.computingStartState)
          if (collectPredicates) {
            this.dfa.predicateVisible = true;
          } else {
            this.dfa.hasPredicateBlockedByAction = true;
          }  
        SemanticContext newSemanticContext = semanticContext;
        if (collectPredicates) {
          int walkAlt = this.dfa.decisionNFAStartState.translateDisplayAltToWalkAlt(alt);
          NFAState altLeftEdge = this.dfa.nfa.grammar.getNFAStateForAltOfDecision(this.dfa.decisionNFAStartState, walkAlt);
          if (!labelContext.isSyntacticPredicate() || p == (altLeftEdge.transition[0]).target)
            newSemanticContext = SemanticContext.and(semanticContext, labelContext); 
        } 
        closure((NFAState)transition0.target, alt, context, newSemanticContext, d, collectPredicates);
      } 
      Transition transition1 = p.transition[1];
      if (transition1 != null && transition1.isEpsilon())
        closure((NFAState)transition1.target, alt, context, semanticContext, d, collectPredicates); 
    } 
  }
  
  public static boolean closureIsBusy(DFAState d, NFAConfiguration proposedNFAConfiguration) {
    return d.closureBusy.contains(proposedNFAConfiguration);
  }
  
  public DFAState reach(DFAState d, Label label) {
    DFAState labelDFATarget = this.dfa.newState();
    List<NFAConfiguration> configs = d.configurationsWithLabeledEdges;
    int numConfigs = configs.size();
    for (int i = 0; i < numConfigs; i++) {
      NFAConfiguration c = configs.get(i);
      if (!c.resolved && !c.resolveWithPredicate) {
        NFAState p = this.dfa.nfa.getState(c.state);
        Transition edge = p.transition[0];
        if (edge != null && c.singleAtomTransitionEmanating) {
          Label edgeLabel = edge.label;
          if (c.context.parent == null || edgeLabel.label != -2)
            if (Label.intersect(label, edgeLabel))
              NFAConfiguration newC = labelDFATarget.addNFAConfiguration((NFAState)edge.target, c.alt, c.context, c.semanticContext);  
        } 
      } 
    } 
    if (labelDFATarget.nfaConfigurations.size() == 0) {
      this.dfa.setState(labelDFATarget.stateNumber, null);
      labelDFATarget = null;
    } 
    return labelDFATarget;
  }
  
  protected void convertToEOTAcceptState(DFAState d) {
    Label eot = new Label(-2);
    int numConfigs = d.nfaConfigurations.size();
    for (int i = 0; i < numConfigs; i++) {
      NFAConfiguration c = d.nfaConfigurations.get(i);
      if (!c.resolved && !c.resolveWithPredicate) {
        NFAState p = this.dfa.nfa.getState(c.state);
        Transition edge = p.transition[0];
        Label edgeLabel = edge.label;
        if (edgeLabel.equals(eot)) {
          d.setAcceptState(true);
          d.nfaConfigurations.clear();
          d.addNFAConfiguration(p, c.alt, c.context, c.semanticContext);
          return;
        } 
      } 
    } 
  }
  
  protected DFAState addDFAStateToWorkList(DFAState d) {
    DFAState existingState = this.dfa.addState(d);
    if (d != existingState) {
      this.dfa.setState(d.stateNumber, existingState);
      return existingState;
    } 
    resolveNonDeterminisms(d);
    int alt = d.getUniquelyPredictedAlt();
    if (alt != -1) {
      d = convertToAcceptState(d, alt);
    } else {
      this.work.add(d);
    } 
    return d;
  }
  
  protected DFAState convertToAcceptState(DFAState d, int alt) {
    if (DFAOptimizer.MERGE_STOP_STATES && d.getNonDeterministicAlts() == null && !d.abortedDueToRecursionOverflow && !d.abortedDueToMultipleRecursiveAlts) {
      DFAState acceptStateForAlt = this.dfa.getAcceptState(alt);
      if (acceptStateForAlt != null) {
        SemanticContext gatedPreds = d.getGatedPredicatesInNFAConfigurations();
        SemanticContext existingStateGatedPreds = acceptStateForAlt.getGatedPredicatesInNFAConfigurations();
        if ((gatedPreds == null && existingStateGatedPreds == null) || (gatedPreds != null && existingStateGatedPreds != null && gatedPreds.equals(existingStateGatedPreds))) {
          this.dfa.setState(d.stateNumber, acceptStateForAlt);
          this.dfa.removeState(d);
          d = acceptStateForAlt;
          return d;
        } 
      } 
    } 
    d.setAcceptState(true);
    this.dfa.setAcceptState(alt, d);
    return d;
  }
  
  public void resolveNonDeterminisms(DFAState d) {
    if (debug)
      System.out.println("resolveNonDeterminisms " + d.toString()); 
    boolean conflictingLexerRules = false;
    Set<Integer> nondeterministicAlts = d.getNonDeterministicAlts();
    if (debug && nondeterministicAlts != null)
      System.out.println("nondet alts=" + nondeterministicAlts); 
    NFAConfiguration anyConfig = d.nfaConfigurations.get(0);
    NFAState anyState = this.dfa.nfa.getState(anyConfig.state);
    if (anyState.isEOTTargetState()) {
      Set<Integer> allAlts = d.getAltSet();
      if (allAlts != null && allAlts.size() > 1) {
        nondeterministicAlts = allAlts;
        if (d.dfa.isTokensRuleDecision()) {
          this.dfa.probe.reportLexerRuleNondeterminism(d, allAlts);
          conflictingLexerRules = true;
        } 
      } 
    } 
    if (!d.abortedDueToRecursionOverflow && nondeterministicAlts == null)
      return; 
    if (!d.abortedDueToRecursionOverflow && !conflictingLexerRules)
      this.dfa.probe.reportNondeterminism(d, nondeterministicAlts); 
    boolean resolved = tryToResolveWithSemanticPredicates(d, nondeterministicAlts);
    if (resolved) {
      if (debug)
        System.out.println("resolved DFA state " + d.stateNumber + " with pred"); 
      d.resolvedWithPredicates = true;
      this.dfa.probe.reportNondeterminismResolvedWithSemanticPredicate(d);
      return;
    } 
    resolveByChoosingFirstAlt(d, nondeterministicAlts);
  }
  
  protected int resolveByChoosingFirstAlt(DFAState d, Set<Integer> nondeterministicAlts) {
    int winningAlt;
    if (this.dfa.isGreedy()) {
      winningAlt = resolveByPickingMinAlt(d, nondeterministicAlts);
    } else {
      int exitAlt = this.dfa.getNumberOfAlts();
      if (nondeterministicAlts.contains(Utils.integer(exitAlt))) {
        winningAlt = resolveByPickingExitAlt(d, nondeterministicAlts);
      } else {
        winningAlt = resolveByPickingMinAlt(d, nondeterministicAlts);
      } 
    } 
    return winningAlt;
  }
  
  protected int resolveByPickingMinAlt(DFAState d, Set<Integer> nondeterministicAlts) {
    int min;
    if (nondeterministicAlts != null) {
      min = getMinAlt(nondeterministicAlts);
    } else {
      min = d.minAltInConfigurations;
    } 
    turnOffOtherAlts(d, min, nondeterministicAlts);
    return min;
  }
  
  protected int resolveByPickingExitAlt(DFAState d, Set<Integer> nondeterministicAlts) {
    int exitAlt = this.dfa.getNumberOfAlts();
    turnOffOtherAlts(d, exitAlt, nondeterministicAlts);
    return exitAlt;
  }
  
  protected static void turnOffOtherAlts(DFAState d, int min, Set<Integer> nondeterministicAlts) {
    int numConfigs = d.nfaConfigurations.size();
    for (int i = 0; i < numConfigs; i++) {
      NFAConfiguration configuration = d.nfaConfigurations.get(i);
      if (configuration.alt != min && (
        nondeterministicAlts == null || nondeterministicAlts.contains(Utils.integer(configuration.alt))))
        configuration.resolved = true; 
    } 
  }
  
  protected static int getMinAlt(Set<Integer> nondeterministicAlts) {
    int min = Integer.MAX_VALUE;
    for (Integer altI : nondeterministicAlts) {
      int alt = altI.intValue();
      if (alt < min)
        min = alt; 
    } 
    return min;
  }
  
  protected boolean tryToResolveWithSemanticPredicates(DFAState d, Set<Integer> nondeterministicAlts) {
    Map<Integer, SemanticContext> altToPredMap = getPredicatesPerNonDeterministicAlt(d, nondeterministicAlts);
    if (altToPredMap.isEmpty())
      return false; 
    this.dfa.probe.reportAltPredicateContext(d, altToPredMap);
    if (nondeterministicAlts.size() - altToPredMap.size() > 1)
      return false; 
    if (altToPredMap.size() == nondeterministicAlts.size() - 1) {
      SemanticContext nakedAltPred;
      BitSet ndSet = BitSet.of(nondeterministicAlts);
      BitSet predSet = BitSet.of(altToPredMap);
      int nakedAlt = ndSet.subtract(predSet).getSingleElement();
      if (nakedAlt == max(nondeterministicAlts)) {
        nakedAltPred = new SemanticContext.TruePredicate();
      } else {
        SemanticContext unionOfPredicatesFromAllAlts = getUnionOfPredicates(altToPredMap);
        if (unionOfPredicatesFromAllAlts.isSyntacticPredicate()) {
          nakedAltPred = new SemanticContext.TruePredicate();
        } else {
          nakedAltPred = SemanticContext.not(unionOfPredicatesFromAllAlts);
        } 
      } 
      altToPredMap.put(Utils.integer(nakedAlt), nakedAltPred);
      int numConfigs = d.nfaConfigurations.size();
      for (int i = 0; i < numConfigs; i++) {
        NFAConfiguration configuration = d.nfaConfigurations.get(i);
        if (configuration.alt == nakedAlt)
          configuration.semanticContext = nakedAltPred; 
      } 
    } 
    if (altToPredMap.size() == nondeterministicAlts.size()) {
      if (d.abortedDueToRecursionOverflow)
        d.dfa.probe.removeRecursiveOverflowState(d); 
      int numConfigs = d.nfaConfigurations.size();
      for (int i = 0; i < numConfigs; i++) {
        NFAConfiguration configuration = d.nfaConfigurations.get(i);
        SemanticContext semCtx = altToPredMap.get(Utils.integer(configuration.alt));
        if (semCtx != null) {
          configuration.resolveWithPredicate = true;
          configuration.semanticContext = semCtx;
          altToPredMap.remove(Utils.integer(configuration.alt));
          if (semCtx.isSyntacticPredicate())
            this.dfa.nfa.grammar.synPredUsedInDFA(this.dfa, semCtx); 
        } else if (nondeterministicAlts.contains(Utils.integer(configuration.alt))) {
          configuration.resolved = true;
        } 
      } 
      return true;
    } 
    return false;
  }
  
  protected Map<Integer, SemanticContext> getPredicatesPerNonDeterministicAlt(DFAState d, Set<Integer> nondeterministicAlts) {
    Map<Integer, SemanticContext> altToPredicateContextMap = new HashMap<Integer, SemanticContext>();
    Map<Integer, OrderedHashSet<SemanticContext>> altToSetOfContextsMap = new HashMap<Integer, OrderedHashSet<SemanticContext>>();
    for (Integer altI : nondeterministicAlts)
      altToSetOfContextsMap.put(altI, new OrderedHashSet<SemanticContext>()); 
    Map<Integer, Set<Token>> altToLocationsReachableWithoutPredicate = new HashMap<Integer, Set<Token>>();
    Set<Integer> nondetAltsWithUncoveredConfiguration = new HashSet<Integer>();
    int numConfigs = d.nfaConfigurations.size();
    for (int i = 0; i < numConfigs; i++) {
      NFAConfiguration configuration = d.nfaConfigurations.get(i);
      Integer altI = Utils.integer(configuration.alt);
      if (nondeterministicAlts.contains(altI))
        if (configuration.semanticContext != SemanticContext.EMPTY_SEMANTIC_CONTEXT) {
          Set<SemanticContext> predSet = altToSetOfContextsMap.get(altI);
          predSet.add(configuration.semanticContext);
        } else {
          nondetAltsWithUncoveredConfiguration.add(altI);
        }  
    } 
    List<Integer> incompletelyCoveredAlts = new ArrayList<Integer>();
    for (Integer altI : nondeterministicAlts) {
      Set<SemanticContext> contextsForThisAlt = altToSetOfContextsMap.get(altI);
      if (nondetAltsWithUncoveredConfiguration.contains(altI)) {
        if (contextsForThisAlt.size() > 0)
          incompletelyCoveredAlts.add(altI); 
        continue;
      } 
      SemanticContext combinedContext = null;
      for (SemanticContext ctx : contextsForThisAlt)
        combinedContext = SemanticContext.or(combinedContext, ctx); 
      altToPredicateContextMap.put(altI, combinedContext);
    } 
    if (incompletelyCoveredAlts.size() > 0) {
      for (int j = 0; j < numConfigs; j++) {
        NFAConfiguration configuration = d.nfaConfigurations.get(j);
        Integer altI = Utils.integer(configuration.alt);
        if (incompletelyCoveredAlts.contains(altI) && configuration.semanticContext == SemanticContext.EMPTY_SEMANTIC_CONTEXT) {
          NFAState s = this.dfa.nfa.getState(configuration.state);
          if (s.incidentEdgeLabel != null && s.incidentEdgeLabel.label != -1)
            if (s.associatedASTNode == null || s.associatedASTNode.token == null) {
              ErrorManager.internalError("no AST/token for nonepsilon target w/o predicate");
            } else {
              Set<Token> locations = altToLocationsReachableWithoutPredicate.get(altI);
              if (locations == null) {
                locations = new HashSet<Token>();
                altToLocationsReachableWithoutPredicate.put(altI, locations);
              } 
              locations.add(s.associatedASTNode.token);
            }  
        } 
      } 
      this.dfa.probe.reportIncompletelyCoveredAlts(d, altToLocationsReachableWithoutPredicate);
    } 
    return altToPredicateContextMap;
  }
  
  protected static SemanticContext getUnionOfPredicates(Map<?, SemanticContext> altToPredMap) {
    SemanticContext unionOfPredicatesFromAllAlts = null;
    Iterator<SemanticContext> iter = altToPredMap.values().iterator();
    while (iter.hasNext()) {
      SemanticContext semCtx = iter.next();
      if (unionOfPredicatesFromAllAlts == null) {
        unionOfPredicatesFromAllAlts = semCtx;
        continue;
      } 
      unionOfPredicatesFromAllAlts = SemanticContext.or(unionOfPredicatesFromAllAlts, semCtx);
    } 
    return unionOfPredicatesFromAllAlts;
  }
  
  protected void addPredicateTransitions(DFAState d) {
    List<NFAConfiguration> configsWithPreds = new ArrayList<NFAConfiguration>();
    int numConfigs = d.nfaConfigurations.size();
    for (int i = 0; i < numConfigs; i++) {
      NFAConfiguration c = d.nfaConfigurations.get(i);
      if (c.resolveWithPredicate)
        configsWithPreds.add(c); 
    } 
    Collections.sort(configsWithPreds, (Comparator<? super NFAConfiguration>)new Object(this));
    List<NFAConfiguration> predConfigsSortedByAlt = configsWithPreds;
    for (int j = 0; j < predConfigsSortedByAlt.size(); j++) {
      NFAConfiguration c = predConfigsSortedByAlt.get(j);
      DFAState predDFATarget = d.dfa.getAcceptState(c.alt);
      if (predDFATarget == null) {
        predDFATarget = this.dfa.newState();
        predDFATarget.addNFAConfiguration(this.dfa.nfa.getState(c.state), c.alt, c.context, c.semanticContext);
        predDFATarget.setAcceptState(true);
        this.dfa.setAcceptState(c.alt, predDFATarget);
        DFAState existingState = this.dfa.addState(predDFATarget);
        if (predDFATarget != existingState) {
          this.dfa.setState(predDFATarget.stateNumber, existingState);
          predDFATarget = existingState;
        } 
      } 
      d.addTransition(predDFATarget, new PredicateLabel(c.semanticContext));
    } 
  }
  
  protected void initContextTrees(int numberOfAlts) {
    this.contextTrees = new NFAContext[numberOfAlts];
    for (int i = 0; i < this.contextTrees.length; i++) {
      int alt = i + 1;
      this.contextTrees[i] = new NFAContext(null, null);
    } 
  }
  
  public static int max(Set<Integer> s) {
    if (s == null)
      return Integer.MIN_VALUE; 
    int i = 0;
    int m = 0;
    for (Integer value : s) {
      i++;
      Integer I = value;
      if (i == 1) {
        m = I.intValue();
        continue;
      } 
      if (I.intValue() > m)
        m = I.intValue(); 
    } 
    return m;
  }
}
